/*
 * vidwin.h
 *
 * Video output window.
 *
 * Portable Windows Library
 *
 * Copyright (c) 2003 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Portable Windows Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: vidwin.h,v $
 * Revision 1.2  2003/05/15 00:52:13  rjongbloed
 * Fixed focus continually going to the video display window.
 *
 * Revision 1.1  2003/03/17 07:41:13  robertj
 * Added classes for video output device that displays in a PWlib window.
 *
 */


#ifndef _VIDWIN_H
#define _VIDWIN_H

#ifdef __GNUC__
#pragma interface
#endif


/**A video stream output window.
 */
class PVideoOutputWindow : public PVideoOutputDevice
{
  PCLASSINFO(PVideoOutputWindow, PVideoOutputDevice);
  public:
  /**@name Construction */
  //@{
    /**Create video output window.
    */
    PVideoOutputWindow();

    /**On destruction, close the associated interactor and clean up the canvas
      */
    ~PVideoOutputWindow();
  //@}

  /**@name Overrides from class PVideoDevice */
  //@{
  private:
    /**Open the device given the device name.
       This is a dummy function and is made private to avoid use.
      */
    virtual BOOL Open(
      const PString & deviceName,   /// Device name to open
      BOOL startImmediate = TRUE    /// Immediately start device
    );

  public:
    /**Determine if the device is currently open.
      */
    virtual BOOL IsOpen();

    /**Close the device.
      */
    virtual BOOL Close();

    /**Get a list of all of the drivers available.
      */
    virtual PStringList GetDeviceNames() const;

    /**Set the colour format to be used.
       Note that this function does not do any conversion. If it returns TRUE
       then the video device does the colour format in native mode.

       To utilise an internal converter use the SetColourFormatConverter()
       function.

       Default behaviour only allows RGB colour formats.
    */
    virtual BOOL SetColourFormat(
      const PString & colourFormat // New colour format for device.
    );

    /**Set the frame size to be used.

       Note that devices may not be able to produce the requested size, and
       this function will fail.  See SetFrameSizeConverter().

       Default behaviour alters the size of the associated interactor.
    */
    virtual BOOL SetFrameSize(
      unsigned width,
      unsigned height
    );

    /**Get the maximum frame size in bytes.

       Note a particular device may be able to provide variable length
       frames (eg motion JPEG) so will be the maximum size of all frames.
      */
    virtual PINDEX GetMaxFrameBytes();

    /**Set a section of the output frame buffer.
      */
    virtual BOOL SetFrameData(
      unsigned x,
      unsigned y,
      unsigned width,
      unsigned height,
      const BYTE * data,
      BOOL endFrame = TRUE
    );

    /**Indicate frame may be displayed.
      */
    virtual BOOL EndFrame();
  //@}

  /**@name new functions */
  //@{
    /**Open the video device using teh specified interactor.
      */
    BOOL Open(
      PInteractor * interactor
    );

    /**Function called when associated interactor needs to redraw itself.
      */
    virtual void OnRedraw(
      PCanvas & canvas   /// Canvas to use when drawing the interactor contents.
    );
  //@}

  protected:
    PInteractor * interactor;
    PPixelImage   image;
    PMutex        mutex;
    BOOL          firstResize;
};


/**A video stream output window as a floating dialog.
 */
class PVideoOutputFloatingDialog : public PFloatingDialog
{
  PCLASSINFO(PVideoOutputFloatingDialog, PFloatingDialog);
  public:
  /**@name Construction */
  //@{
    /**Create video output window.
       Note that the the PVideoOutputWindow instance is expected to be "owned"
       by something else and its memory deallocated by that other entity.
       Before it does so, however, it should call CloseDevice() first.
    */
    PVideoOutputFloatingDialog(
      PInteractor * parent,
      PVideoOutputWindow * device
    );
  //@}

  /**@name Overrides from class PDialog */
  //@{
   /**The system calls this whenever the window is being closed. There is no
      way to prevent this action at this point. Override the \Ref{OnCancel()}
      function if you wish to control if the close operation proceeds.

       The application writer may override this function to do any clean up
       or state saving operations.
       
       The default action does nothing.
     */
    virtual void OnClose();

   /**Function called when the dialog Cancel button has been pressed. The
       default behaviour is to close the dialog using \Ref{Close()}.
     */
    virtual void OnCancel();

   /**The system calls this whenever it requires that the interactors usable
       area needs to be updated and redrawn.

       The canvas is usually a \Ref{PRedrawCanvas} instance with its clip
       region set to the current update region as determined by the
       \Ref{Invalidate()} and \Ref{Validate()} functions as well as those
       areas that were obscured by overlapping windows.
     */
    virtual void OnRedraw(
      PCanvas & canvas   /// Canvas to use when drawing the interactor contents.
    );
  //@}

  /**@name new functions */
  //@{
    /**Close the device.
      */
    void CloseDevice();

    /**Get the currently attached vide device.
      */
    PVideoOutputWindow * GetDevice() const { return device; }
  //@}

  protected:

    PVideoOutputWindow * device;
    PMutex               mutex;
};


#endif // _VIDWIN_H


// End Of File ///////////////////////////////////////////////////////////////
