/*
 * vidwin.cxx
 *
 * Video output window.
 *
 * Portable Windows Library
 *
 * Copyright (c) 2003 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Portable Windows Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: vidwin.cxx,v $
 * Revision 1.3  2003/05/15 00:52:47  rjongbloed
 * Fixed focus continually going to the video display window.
 *
 * Revision 1.2  2003/05/14 07:52:36  rjongbloed
 * Fixed some trace logs.
 *
 * Revision 1.1  2003/03/17 07:40:58  robertj
 * Added classes for video output device that displays in a PWlib window.
 *
 */

#ifdef __GNUC__
#pragma implementation "vidwin.h"
#endif

#include <pwlib.h>

#include <pwclib/vidwin.h>
#include <ptlib/vconvert.h>


#define new PNEW


///////////////////////////////////////////////////////////////

PVideoOutputWindow::PVideoOutputWindow()
{
  preferredColourFormat = "RGB24";
  SetColourFormat(preferredColourFormat);

  interactor = NULL;
  firstResize = TRUE;
}


PVideoOutputWindow::~PVideoOutputWindow()
{
  Close();
}


BOOL PVideoOutputWindow::Open(const PString & /*deviceName*/,
                              BOOL /*startImmediate*/)
{
  return IsOpen();
}


BOOL PVideoOutputWindow::Open(PInteractor * i)
{
  PWaitAndSignal wait(mutex);
  delete interactor;
  interactor = PAssertNULL(i);
  return interactor != NULL;
}


BOOL PVideoOutputWindow::IsOpen()
{
  PWaitAndSignal wait(mutex);
  return interactor != NULL;
}


BOOL PVideoOutputWindow::Close()
{
  mutex.Wait();
  PInteractor * i = interactor;
  interactor = NULL;
  mutex.Signal();

  if (i != NULL)
    i->Close();

  return TRUE;
}


PStringList PVideoOutputWindow::GetDeviceNames() const
{
  PStringList list;
  list += "PVideoOutputFloatingDialog";
  return list;
}


BOOL PVideoOutputWindow::SetColourFormat(const PString & colourFormat)
{
  if (colourFormat != "RGB24" && colourFormat != "RGB32")
    return FALSE;

  return PVideoOutputDevice::SetColourFormat(colourFormat);
}


BOOL PVideoOutputWindow::SetFrameSize(unsigned width, unsigned height)
{
  PWaitAndSignal wait(mutex);

  if (interactor == NULL)
    return FALSE;

  if (image.IsNULL() ||
      image->Width() != (PDIMENSION)width ||
      image->Height() != (PDIMENSION)height) {

    unsigned bitsPerPixel = 32;
    if (colourFormat == "RGB24")
      bitsPerPixel = 24;

    PTRACE(3, "PWLib\tNew video output to window size set: "
           << width << 'x' << height << 'x' << bitsPerPixel << " bit");

    image = PPixelImage(width, height, bitsPerPixel);
  }

  if (firstResize || frameWidth != width || frameHeight != height) {
    interactor->SetDimensions(width, height, PInteractor::PixelCoords);
    interactor->Show();
    firstResize = FALSE;
  }

  return PVideoOutputDevice::SetFrameSize(width, height);
}


PINDEX PVideoOutputWindow::GetMaxFrameBytes()
{
  PWaitAndSignal wait(mutex);

  if (image.IsNULL())
    return 0;

  return image->GetPixelDataSize();
}


BOOL PVideoOutputWindow::SetFrameData(unsigned x, unsigned y,
                                      unsigned width, unsigned height,
                                      const BYTE * data, BOOL endFrame)
{
  PWaitAndSignal wait(mutex);

  if (interactor == NULL || image.IsNULL())
    return FALSE;

  if (x == 0 && y == 0 && width == frameWidth && height == frameHeight) {
    if (converter != NULL)
      converter->Convert(data, image->GetPixelDataPtr());
    else
      memcpy(image->GetPixelDataPtr(), data, image->GetPixelDataSize());
  }
  else {
    if (converter != NULL)
      return FALSE;
    for (unsigned dy = 0; dy < height; dy++)
      image->SetRaster(x, y+dy, data+dy*width*3, width);
  }

  if (endFrame)
    return EndFrame();

  return TRUE;
}


BOOL PVideoOutputWindow::EndFrame()
{
  PWaitAndSignal wait(mutex);

  if (interactor == NULL || image.IsNULL())
    return FALSE;

  interactor->Invalidate();
  return TRUE;
}


void PVideoOutputWindow::OnRedraw(PCanvas & redrawCanvas)
{
  if (!image.IsNULL())
    redrawCanvas.DrawPixels(0, 0, image);
}


///////////////////////////////////////////////////////////////

PVideoOutputFloatingDialog::PVideoOutputFloatingDialog(PInteractor * parent,
                                                       PVideoOutputWindow * dev)
  : PFloatingDialog(parent),
    device(dev)
{
}


void PVideoOutputFloatingDialog::OnClose()
{
  CloseDevice();
  PFloatingDialog::OnClose();
}


void PVideoOutputFloatingDialog::OnCancel()
{
  CloseDevice();
  Close();
}


void PVideoOutputFloatingDialog::OnRedraw(PCanvas & canvas)
{
  mutex.Wait();
  if (device != NULL)
    device->OnRedraw(canvas);
  mutex.Signal();
}


void PVideoOutputFloatingDialog::CloseDevice()
{
  mutex.Wait();

  if (device != NULL) {
    device->Close();
    device = NULL;
  }

  mutex.Signal();
}



// End Of File ///////////////////////////////////////////////////////////////
